/* Datastructures for the meme machine. */


/*****************************************************************************/
/* tunable parameters referred to throughout the program                     */

/* RAM is the no. of bytes of physical RAM you have available. You can make this
   number bigger than the memory you have, but your machine will thrash a lot.
   This value is used to regulate calls to dream() & defrag() to clear space */
   
#define RAM			128000000

/* MAX_DATA_SIZE is the absolute maximum size of a 'net' construct, ideally
   500,000,000  (i.e. 2 gigabytes/4) for a 32-bit Linux platform.
   It's smaller here so that the code is more portable.
   Note that because of file-holes, the 'net' file can be apparently larger
   than filesystem space available.
   Make this smaller to make your 'net' file smaller/more portable, but try
   to keep it large to allow expandability if you get more RAM.             */
   
#define MAX_DATA_SIZE		50000000

/* MAX_IPOP is the maximum number of outputs a node can have. On average a
   node has 100 outputs. */

#define MAX_IPOP		80000

/* MAX_LEVELS is the highest level a node can reach. Since a level 25 node
   can represent the entire text of 'Alice in Wonderland', 100 is a good
   standard here */
					  					  
#define MAX_LEVELS		100	/* max level of a node*/

/* PRUNE (verb - to cut away unwanted material, not the noun, the dried fruit)
  This defines the number of levels up that a search will be pruned. Usefull
  values are 2-10. Higher=more interpretations searched, but slower speed */
  
#define PRUNE 			6

/* TWEAK_PRUNE - similar to PRUNE, but used with in tweak-based searching.
  Should generally be the same as or smalled than PRUNE */
  
#define TWEAK_PRUNE		2

/* WAKE_INCREASE - defines how quickly the frequency of calls to  dream()
   tails off. Higher
   values mean more dreaming early on, longer learning times but less bias
   being passed up from low-level items to higher-level items.
   Useful values=20-99 */
   
#define WAKE_INCREASE		70

/* MAX_WAKE is the maximum number of nodes that can be made before a 
   call to dream(). */

#define MAX_WAKE 1000000

/* SCRATCH_SIZE is used for the scratch area, where a node can get written out
   for comparision with another node. */

#define SCRATCH_SIZE		200000

#define MAX_TWEAKS		50    /*max. number of tweaks allowed at once*/

#define MAX_WMSIZE		4    /* max working-memory size for tweaks */

#define ABS_MAX_WMSIZE		20   /* space allocated for working memory */

#define FRAG_SIZE		1000 /* used in defrag() - relats to L1cache*/

#define MAX_QUOTED_LINES	30   /*max number of lines of quote in a text*/

#define NUM_BASE_NODES		128   	/*number of base nodes in system */
					/* (i.e. 128 ASCII characters) */
					
#define BASE_NODE_SPACE		1024   /*net space used by bases (128 x 8)*/

/* #define DOS */		/* uncomment this out if compiling on DOS */

/*----------------------------------------------------------------------*/
/* you probably dont want to play with anything below here...           */

#define SPEC			0	/* spec-int, each byte has meaning */
					/* byte 2= first, byte 3/4 = level */
					/* top bit = responce boolean */
#define FIRST_MASK		65535     /* to mask out the bottom 2 bytes */
#define LEVEL_MASK		16711680   /* 255x256x256 - byte 2 of 4*/
#define LEVEL_DIV		65536

	   
#define LHS_IP			1       /* LHS input of node */
#define RHS_IP			2	/* RHS input of node */
#define LHS_OP			3	/* pointer to LHS output-list */
#define RHS_OP			4	/* pointer to RHS output-list */
#define NUM_LHS_OP		5	/*count of LHS outputs */
#define NUM_RHS_OP		6	/* count of RHS outputs */
#define SPARE			7       /* for future use */
#define TOP			8	/* top of node area */


/*****************************************************************************
 * Usefull constants 
 */

#define HIGH_BIT                1073741824 /*integer with 2nd highest bit set*/

#define TRUE			1
#define	FALSE			0 

#define LHS			0
#define RHS			1


/******************************************************************************
 * nodelist is just a string of nodes. used to store text read in 
 *  and output nodelists for the program
 */
 
typedef struct
{
   int *data; /* contents of input/output text. */
   int size;
} nodelist_type;

/******************************************************************************
 * net = the major data-type of the system, used to store the 
 * virtual knowledge-net, the basis of the meme machine.
 */

typedef struct
{
   int ptr;  /*pointer to next free space when allocating chunks */
   int nodespace; /* size (in ints) of space used by node area*/
   int actual_num_nodes; /* actual number of nodes being used in net*/
   int deleted_node;  /*pointer to top of linked-list of deleted nodes. */
   int next_dream;
   int mode;    /*This is: 1 (reading), 2 (listen), 3 (speak) or 4 (prompt) */
   int chatlog;  /*whether an input text is chat-logs (1) or story (0) */
   int trace;  /*used to show diagnostics when running the program */
   int result; /* set in compare() - either 0 (less than) or 1 (greater than)*/
   int scratch[SCRATCH_SIZE]; int scratch_posn; int scratch_size;
   int free[MAX_IPOP];  /*linkedlists of locations of free chunks of all sizes*/
   
   int start,size;
   int best,  bestlevel, bestnext;
   
   int tweaked, tweak_lhs, tweak_rhs, tweaklevel, tweak_posn,
       tweak_b4;
      
   /* used for tweaking.... */   
   int  besttweaked,
       best_tweak_lhs, best_tweak_rhs, best_tweaklevel, best_tweakposn,
       best_tweakb4, template, actual;
    int metabest;
 
    int spoken;  /* count of chars in a responce... */
    
    int wm[ABS_MAX_WMSIZE];  /*working memory,for exotic multi-tweak matching */
    int wmsize;
    
    int best_wm[ABS_MAX_WMSIZE];
    int best_wmsize;  
    
   int data[MAX_DATA_SIZE];
} net_type;

/******************************************************************************
 * The 'tweaks' type is used to store a collection of tweaks
 */
 
typedef struct
{
   int lhs[MAX_TWEAKS];
   int rhs[MAX_TWEAKS];
   int b4[MAX_TWEAKS];
   int posn[MAX_TWEAKS];
   int current,total;
} tweaks_type;

/******************************************************************************
 * The 'table' type is used in the shuffle_defrag() routine to store 
 * an ordered list of chunks
 */

typedef struct
{
   int posn[FRAG_SIZE+2];  /* +1 for fencepost, +1 for copying from (x+1) */
   int node[FRAG_SIZE+2];
   int  size;
} table_type;

/****************************************************************************
 * Function declarations (for clean compiles)
 */
/* init.c */
int responce_node(int node);

/* main.c */
void need_input(net_type *net, nodelist_type *input);
int off_the_end(net_type *net, int node, nodelist_type *input, int posn);
void find_match(net_type *net, int node, nodelist_type *input, int posn);

/* high.c */
void print_node(net_type *net, int node);
int get_level(net_type *net, int node);
int find_node(net_type *net, int lhs, int rhs);
int get_node_size(net_type *net, int node);
void print_node_to_file(net_type *net, int node, FILE *fp_out);
void dream(net_type *net);

/* init.c */
void defrag(net_type *net);
void initialise();

/* mid.c */
void delete_node(net_type *net,int node);
int new_node(net_type *net);
int make_node(net_type *net, int lhs, int rhs);

/* bottom.c */
int get_chunk(net_type *net, int size);
void free_chunk(net_type *net, int start, int size);
void dump_stats(net_type *net);
void read_text(net_type *net, nodelist_type *input, nodelist_type *output,
	       char *filename);

/* low.c */
void add_lhs_output(net_type *net, int node, int output);
void add_rhs_output(net_type *net, int node, int output);
void del_lhs_output(net_type *net, int node, int output);
void del_rhs_output(net_type *net, int node, int output);

/* tweak.c */
int compare(net_type *net, int node, nodelist_type *input, int posn);
int get_meta(net_type *net, nodelist_type *input, int posn, int meta_rhs);
void find_tweak_match(net_type *net, int node, nodelist_type *input, int posn);
int tweak_build(net_type *net, int node, nodelist_type *input, int lhs_posn,
	        tweaks_type *tweaks);
void multi_tweak_match(net_type *net, int node, nodelist_type *input, int posn,
                       tweaks_type *tweaks);
