/* highlevel.c - high level constructs for the meme machine 
 * 
 * functions defined here are:
 * find_node()
 * print_node()
 * print_node_to_file()
 * add_to_input()
 * get_node_size()
 * get_level()
 * dream()
 */

#include <stdio.h>
#include <sys/mman.h>
#include "datastru.h"


/**************************************************************************
 * int find_node(net, node1,node2)
 * This will return the number of a node built up of lhs and rhs.
 * if none exists, return's 0. 
 *
 */

int find_node(net_type *net, int lhs, int rhs)
{
   int metanode;
   int outputs;
   int start, end;
   int num_lhs_op, num_rhs_op;

   /*quick bomb-out - if lhs or rhs has no outputs, a
     match is impossible, so waste no more time! */
   if (net->data[lhs+NUM_LHS_OP]==0) return(0);
   if (net->data[rhs+NUM_RHS_OP]==0) return(0);
   
   /*now look through the outputs of nodeX for a node built of lhs & rhs*/
   /* Firstly, see what will be the quickest output list to search...*/
   
   num_lhs_op=net->data[lhs+NUM_LHS_OP];
   num_rhs_op=net->data[rhs+NUM_RHS_OP];
   if (num_lhs_op<num_rhs_op)
   { start=net->data[lhs+LHS_OP]; end=start+num_lhs_op; }
   else
   { start=net->data[rhs+RHS_OP]; end=start+num_rhs_op;}
   
   for (outputs=start; outputs<end; outputs++)
   {
     metanode=net->data[outputs]; /*output node to check*/
     
     /*Check that the 2 inputs of 'node' match our 2 requests. */
     if (net->data[metanode+LHS_IP]==lhs &&
         net->data[metanode+RHS_IP]==rhs )
     {
       return(metanode);
     }
   }
   return(0);   /* no match found... :(  */
}



/**************************************************************************/
/* print_node() - recursive - print out what a node is made up of.
 */
void print_node(net_type *net, int node)
{

  /* 2 conditions to handle... 1) node<=NUM_BASE_NODES (just print char) and
                               2) node>NUM_BASE_NODES (recurse!) */

  /* firstly, sanity check*/
  if (node==0) return; 

  if (node<=BASE_NODE_SPACE) printf("%c",node/TOP);
  else
  {	       
      print_node(net, net->data[node+LHS_IP]);
      print_node(net, net->data[node+RHS_IP]);     
  }
}

/**************************************************************************/
/* print_node_to_file() - recursive - print out what a node is made up of.
 * except output to a file.
 */
void print_node_to_file(net_type *net, int node, FILE *fp_out)
{

  /* 2 conditions to handle... 1) node<=NUM_BASE_NODES (just print char) and
                               2) node>NUM_BASE_NODES (recurse!) */

  /* firstly, sanity check*/
  if (node==0); 

  if (node<=BASE_NODE_SPACE)  fprintf(fp_out,"%c",node/TOP);
  else
  {     
    print_node_to_file(net,net->data[node+LHS_IP] ,fp_out);
    print_node_to_file(net,net->data[node+RHS_IP] ,fp_out);
  }
}



/**************************************************************************/
/* add_to_input() - recursive - adds a node, char by char, to end of input
 */
void add_to_input(net_type *net, nodelist_type *input, int node)
{

  /* 2 conditions to handle... 1) node<=NUM_BASE_NODES (just print char) and
                               2) node>NUM_BASE_NODES (recurse!) */
  /* printf("print_node(net,%i)\n",node); */
  if (node<=BASE_NODE_SPACE)
  {
     input->size++; input->data[input->size]=node/TOP;
  }
  else
  {	        
    add_to_input(net, input, net->data[node+LHS_IP]);
    add_to_input(net, input, net->data[node+RHS_IP]);    
  }
}

/*************************************************************************
 * get_node_size() - recursive routine to count base input nodes in a node
 */

int get_node_size(net_type *net, int node)
{
  int length;
  
  if (node<BASE_NODE_SPACE) return(1);
  else
  {
   length=0;
   length=length+get_node_size(net,net->data[node+LHS_IP]); 
   length=length+get_node_size(net,net->data[node+RHS_IP]);
   return(length);
  }
}

/*****************************************************************************
 * get_level(net,node);
 * return the level of a node
 */
 
int get_level(net_type *net, int node)
{
   int level;
   
   level=(net->data[node] & LEVEL_MASK) / LEVEL_DIV;
   return(level);
}

/*****************************************************************************
 *     dream(net) - have a clear-out of some nodes with no outputs.
 *                  This is much like simulated anealing in Neural Nets.
 *		    Called from make_node()
 */

void dream(net_type *net)
{
   int node;
   int total;
   int trace, tracetotal, traceposn;

   if (net->trace>=1)
       printf("************************ Dream ***************************\n"); 
   /*OK, we go through the net->node list, check num_outputs for each
     node, delete the ones with no outputs */
     
   total=0;
   /* dream() works in 2 stages now, marking nodes for deletion, then
      deleting them. Marking is done bot->top, deleting top->bot */
   /* delete nodes that a) have no outputs and
                        b) are not already deleted. (i.e. do have inputs) */
   for (node=BASE_NODE_SPACE+TOP; node<net->nodespace; node+=TOP)
   {
      if (net->data[node+NUM_LHS_OP]==0 && net->data[node+NUM_RHS_OP]==0 &&
          net->data[node+LHS_IP]!=0) 
      {       
        total++;
        net->data[node+SPARE]=1; /*mark for deletion*/
      }
      else net->data[node+SPARE]=0;
   }
   for (node=net->nodespace-TOP; node>BASE_NODE_SPACE+TOP; node-=TOP)
   { 	   
      if (net->data[node+SPARE]==1)
         delete_node(net,node);	
   }
   
   /* good oppertunity to sync the network back to disk here... */
   /* msync(net,sizeof(net_type), MS_ASYNC); */
   
   if (net->trace>=1)
     printf("*** deleted %i nodes in dream, %i left, nodespace=%i *****\n",
             total,net->actual_num_nodes,net->nodespace);
   if (total<5)
   {
     printf("WOA! network is too small :( \n");
     printf("trace of free spaces follows... \n");
     for (trace=1; trace<MAX_IPOP; trace++)
	 {
	    traceposn=net->free[trace]; tracetotal=0;
	    while(traceposn!=0) 
	       {traceposn=net->data[traceposn]; tracetotal++;}
	    if (tracetotal!=0)
	      printf("size %i, free=%i\n",trace,tracetotal);
	 }
      exit(0);
   }
   /* define when the next dream() will be... */
   if (net->next_dream<MAX_WAKE)
   net->next_dream=net->next_dream*100/WAKE_INCREASE; 
}

